/* Copyright (c) 2008-2022 ZAO "MCST". All rights reserved. */

#ifndef _E2KINTRIN_H
#define _E2KINTRIN_H

#ifdef __cplusplus
extern "C" {
#endif

/* *****************************************************************************
 *
 * В этом файле описаны интринсики на основные операции системы команд процессоров
 * Эльбрус, которые могут понадобиться для оптимизации или низкоуровневого программирования.
 * Каждый интринсик описывает одну команду Эльбруса. Для параметров и результатов
 * используются целые типы данных: unsigned int для данных формата 32, unsigned long long
 * для данных формата 64 и вектор из 2-х unsigned long long для формата 128.
 * За деталями работы операций см. систему команд.
 *
 * Примеры использования:
 *
 * #include <e2kintrin.h>
 *
 * unsigned src1;
 * unsigned dst = __builtin_e2k_scls (src1, 5);
 *
 * __di src2;
 * __di dst = __builtin_e2k_paddb (0x0102030405060708LL, src2);
 *
 * __v2di qsrc1;
 * __v2di qdst = __builtin_e2k_qpaddb (qsrc1, (__v2di) {0x0102030405060708LL, 0x090a0b0c0d0e0f00LL});
 *
 * *****************************************************************************/

/* Внутренние типы для реализации интринсиков */

typedef unsigned long long __di;

typedef char __v8qi __attribute__ ((__vector_size__ (8)));
typedef short __v4hi __attribute__ ((__vector_size__ (8)));
typedef int __v2si __attribute__ ((__vector_size__ (8)));
typedef long long __v1di __attribute__ ((__vector_size__ (8)));
typedef float __v2sf __attribute__ ((__vector_size__ (8)));

typedef char __v16qi __attribute__ ((__vector_size__ (16)));
typedef signed char __v16qs __attribute__ ((__vector_size__ (16)));
typedef unsigned char __v16qu __attribute__ ((__vector_size__ (16)));
typedef short __v8hi __attribute__ ((__vector_size__ (16)));
typedef unsigned short __v8hu __attribute__ ((__vector_size__ (16)));
typedef int __v4si __attribute__ ((__vector_size__ (16)));
typedef unsigned int __v4su __attribute__ ((__vector_size__ (16)));
typedef long long __v2di __attribute__ ((__vector_size__ (16)));
typedef unsigned long long __v2du __attribute__ ((__vector_size__ (16)));
typedef float __v4sf __attribute__ ((__vector_size__ (16)));
typedef double __v2df __attribute__ ((__vector_size__ (16)));

typedef union {
    int i;
    float f;
} __attribute__ ((__may_alias__)) type_union_32;

typedef union {
    double d0;
    unsigned long long l0;
    struct {
        int i0, i1;
    } i;
    struct {
        float f0, f1;
    } f;
    char __v8qi __attribute__ ((__vector_size__ (8)));
    short __v4hi __attribute__ ((__vector_size__ (8)));
    int __v2si __attribute__ ((__vector_size__ (8)));
    long long __v1di __attribute__ ((__vector_size__ (8)));
    float __v2sf __attribute__ ((__vector_size__ (8)));
} __attribute__ ((__may_alias__)) type_union_64;

typedef union {
    struct {
        double d0, d1;
    } d;
    struct {
        unsigned long long l0, l1;
    } l;
    struct {
        int i0, i1, i2, i3;
    } i;
    struct {
        float f0, f1, f2, f3;
    } f;
    struct {
        short h0, h1, h2, h3, h4, h5, h6, h7;
    } h;
    struct {
        char c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13, c14, c15;
    } c;
    char __v16qi __attribute__ ((__vector_size__ (16)));
    signed char __v16qs __attribute__ ((__vector_size__ (16)));
    short __v8hi __attribute__ ((__vector_size__ (16)));
    int __v4si __attribute__ ((__vector_size__ (16)));
    long long __v2di __attribute__ ((__vector_size__ (16)));
    float __v4sf __attribute__ ((__vector_size__ (16)));
    double __v2df __attribute__ ((__vector_size__ (16)));
} __attribute__ ((__may_alias__)) type_union_128;


#ifndef __E2K_INLINE
#define __E2K_INLINE static __inline
# endif  /* __E2K_INLINE */

/* *****************************************************************************
 *           Реализация оставшихся неподдержанными операций Эльбрус
 *  (все остальные, закомментированные в данном файле, встроены в компилятор)
 * *****************************************************************************/

#define TWO_OPERANDS_INSTRUCTIONS(instr)                     \
__E2K_INLINE __di __attribute__((__always_inline__))         \
__builtin_e2k_##instr (__di src1, __di src2)                 \
{                                                            \
    __di dst;                                                \
_Pragma ("asm_inline")                                       \
    asm (""#instr" %[src1_name], %[src2_name], %[dst_name]"  \
        : [dst_name] "=r" (dst)                              \
        : [src1_name] "rI" (src1), [src2_name] "ri" (src2)); \
    return dst;                                              \
}

#define TWO_OPERANDS_INSTRUCTIONS_QSQ(instr)                 \
__E2K_INLINE __v2di __attribute__((__always_inline__))       \
__builtin_e2k_##instr (__v2di src1, unsigned int src2)       \
{                                                            \
    __v2di dst;                                              \
_Pragma ("asm_inline")                                       \
    asm (""#instr" %[src1_name], %[src2_name], %[dst_name]"  \
        : [dst_name] "=r" (dst)                              \
        : [src1_name] "rI" (src1), [src2_name] "ri" (src2)); \
    return dst;                                              \
}

#define __GETPFPFR(cw) _Pragma ("asm_inline") __asm__ __volatile__ ("rrs %%pfpfr, %0" : "=r" (cw))
#define __SETPFPFR(cw) _Pragma ("asm_inline") __asm__ __volatile__ ("rws %0, %%pfpfr" : : "r" (cw))

/* *****************************************************************************
 *                         Сложение и вычитание по модулю
 * *****************************************************************************/

/* сложение 8-ми байт (8S/8U)
 * __di __builtin_e2k_paddb (__di, __di) */

/* сложение 4-х шортов (16S/16U)
 * __di __builtin_e2k_paddh (__di, __di) */

/* сложение 2-х интов (32S/32U)
 * __di __builtin_e2k_paddw (__di, __di) */

/* сложение 1-го лонга (64S/64U)
 * __di __builtin_e2k_paddd (__di, __di) */

/* вычитание 8-ми байт (8S/8U)
 * __di __builtin_e2k_psubb (__di, __di) */

/* вычитание 4-х шортов (16S/16U)
 * __di __builtin_e2k_psubh (__di, __di) */

/* вычитание 2-х интов (32S/32U)
 * __di __builtin_e2k_psubw (__di, __di) */

/* вычитание 1-го лонга (64S/64U)
 * __di __builtin_e2k_psubd (__di, __di) */

/* *****************************************************************************
 *                         Сложение и вычитание c насыщением
 * *****************************************************************************/

/* сложение 8-ми знаковых байт (8S)
 * __di __builtin_e2k_paddsb (__di, __di) */

/* сложение 4-х знаковых шортов (16S)
 * __di __builtin_e2k_paddsh (__di, __di) */

/* сложение 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_paddusb (__di, __di) */

/* сложение 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_paddush (__di, __di) */

/* вычитание 8-ми знаковых байт (8S)
 * __di __builtin_e2k_psubsb (__di, __di) */

/* вычитание 4-х знаковых шортов (16S)
 * __di __builtin_e2k_psubsh (__di, __di) */

/* вычитание 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_psubusb (__di, __di) */

/* вычитание 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_psubush (__di, __di) */

/* *****************************************************************************
 *                         Среднее арифметическое
 * *****************************************************************************/

/* среднее 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_pavgusb (__di, __di) */

/* среднее 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_pavgush (__di, __di) */

/* *****************************************************************************
 *                             Максимум и минимум
 * *****************************************************************************/

/* максимум 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_pmaxub (__di, __di) */

/* максимум 4-х знаковых шортов (16S)
 * __di __builtin_e2k_pmaxsh (__di, __di) */

/* минимум 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_pminub (__di, __di) */

/* минимум 4-х знаковых шортов (16S)
 * __di __builtin_e2k_pminsh (__di, __di) */

/* *****************************************************************************
 *                             Умножение шортов
 * *****************************************************************************/

/* умножение 4-х беззнаковых шортов (16U) и выдача старших 16 разрядов произведений
 * __di __builtin_e2k_pmulhuh (__di, __di) */

/* умножение 4-х знаковых шортов (16S) и выдача старших 16 разрядов произведений
 * __di __builtin_e2k_pmulhh (__di, __di) */

/* умножение 4-х шортов (16S/16U) и выдача младших 16 разрядов произведений
 * __di __builtin_e2k_pmullh (__di, __di) */

/* *****************************************************************************
 *                             Умножение интов
 * *****************************************************************************/

/* умножение беззнаковых интов (32U) и получение 64-х разрядного произведения
 * __di __builtin_e2k_umulx (unsigned, unsigned) */

/* умножение знаковых интов (32S) и получение 64-х разрядного произведения
 * __di __builtin_e2k_smulx (unsigned, unsigned) */

/* *****************************************************************************
 *                 Умножение шортов с горизонтальным сложением
 * *****************************************************************************/

/* умножение 4-х знаковых шортов (16S) и попарное сложение смежных интовых произведений
 * __di __builtin_e2k_pmaddh (__di, __di) */

/* *****************************************************************************
 *                Cложение абсолютных разностей беззнаковых байт
 * *****************************************************************************/

/* cложение абсолютных разностей 8-ми беззнаковых байт (8U)
 * __di __builtin_e2k_psadbw (__di, __di) */

/* *****************************************************************************
 *                                Cравнение
 * *****************************************************************************/

/* сравнение на "равно" 8-ми байт (8S/8U)
 * __di __builtin_e2k_pcmpeqb (__di, __di) */

/* сравнение на "равно" 4-х шортов (16S/16U)
 * __di __builtin_e2k_pcmpeqh (__di, __di) */

/* сравнение на "равно" 2-х интов (32S/32U)
 * __di __builtin_e2k_pcmpeqw (__di, __di) */

/* сравнение на "больше" 8-ми знаковых байт (8S)
 * __di __builtin_e2k_pcmpgtb (__di, __di) */

/* сравнение на "больше" 4-х знаковых шортов (16S)
 * __di __builtin_e2k_pcmpgth (__di, __di) */

/* сравнение на "больше" 2-х знаковых интов (32S)
 * __di __builtin_e2k_pcmpgtw (__di, __di) */

/* *****************************************************************************
 *                           Логические операции
 * *****************************************************************************/

/* поразрядное логическое AND 64-х разрядов (64S/64U)
 * __di __builtin_e2k_pandd (__di, __di) */

/* поразрядное логическое AND 64-х разрядов c инверсией второго операнда (64S/64U)
 * __di __builtin_e2k_pandnd (__di, __di) */

/* поразрядное логическое OR 64-х разрядов (64S/64U)
 * __di __builtin_e2k_pord (__di, __di) */

/* поразрядное логическое XOR 64-х разрядов (64S/64U)
 * __di __builtin_e2k_pxord (__di, __di) */

/* *****************************************************************************
 *                             Формирование маски
 * *****************************************************************************/

/* формирование 16-разрядной маски из знаков каждого байта (8S)
 * __di __builtin_e2k_pmovmskb (__di, __di) */

/* формирование 4-разрядной маски из знаков каждого инта (32S)
 * __di __builtin_e2k_pmovmskps (__di, __di) */

/* формирование 2-разрядной маски из знаков каждого лонга (64S)
 * __di __builtin_e2k_pmovmskpd (__di, __di) */

/* *****************************************************************************
 *                               Битовый сдвиг
 * *****************************************************************************/

/* сдвиг влево 4-х шортов (16S/16U)
 * __di __builtin_e2k_psllh (__di, __di) */

/* сдвиг влево 2-х интов (32S/32U)
 * __di __builtin_e2k_psllw (__di, __di) */

/* сдвиг влево лонга (64S/64U)
 * __di __builtin_e2k_pslld (__di, __di) */

/* логический сдвиг вправо 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_psrlh (__di, __di) */

/* логический сдвиг вправо 2-х беззнаковых интов (32U)
 * __di __builtin_e2k_psrlw (__di, __di) */

/* логический сдвиг вправо беззнакового лонга (64U)
 * __di __builtin_e2k_psrld (__di, __di) */

/* арифметический сдвиг вправо 4-х знаковых шортов (16S)
 * __di __builtin_e2k_psrah (__di, __di) */

/* арифметический сдвиг вправо 2-х знаковых интов (32S)
 * __di __builtin_e2k_psraw (__di, __di) */

/* *****************************************************************************
 *                              Байтовый сдвиг
 * *****************************************************************************/

/* сдвиг влево 16-ти байт (сцепленные первые 2 операнда) на число байт,
 * равное константе в 3-м операнде, и выдача старших 8 байт
 * __di __builtin_e2k_psllqh (__di, __di, imm8) */

/* сдвиг влево 16-ти байт (сцепленные первые 2 операнда) на число байт,
 * равное константе в 3-м операнде, и выдача младших 8 байт
 * __di __builtin_e2k_psllql (__di, __di, imm8) */

/* логический сдвиг вправо 16-ти байт (сцепленные первые 2 операнда) на число байт,
 * равное константе в 3-м операнде, и выдача старших 8 байт
 * __di __builtin_e2k_psrlqh (__di, __di, imm8) */

/* логический сдвиг вправо 16-ти байт (сцепленные первые 2 операнда) на число байт,
 * равное константе в 3-м операнде, и выдача младших 8 байт
 * __di __builtin_e2k_psrlql (__di, __di, imm8) */

/* *****************************************************************************
 *                          Вставка / выделение шорта
 * *****************************************************************************/

/* выделение беззнакового шорта из 8-ми шортов (сцепленные первые 2 операнда) с номером,
 * задаваемым константой в 3-м операнде (ноль соответствует младшему шорту)
 * __di __builtin_e2k_pextrh (__di, __di, imm8) */

/* вcтавка шорта из младшей части операнда 2 в операнд 1 на место с номером,
 * задаваемым константой в 3-м операнде (ноль соответствует младшему шорту)
 * __di __builtin_e2k_pinsh (__di, __di, imm8) */

/* *****************************************************************************
 *                                 Перестановка
 * *****************************************************************************/

/* выделение из 4-х входных интов двух, положение которых определяется 4-мя
 * младшими разрядами константы 3-го операнда
 * __di __builtin_e2k_pshufw (__di, __di, imm8) */

/* перестановка 4-х шортов в соответствии с порядком, определенным в 8-ми младших
 * разрядах константы 2-го операнда
 * __di __builtin_e2k_pshufh (__di, imm8) */

/* *****************************************************************************
 *                                 Упаковка
 * *****************************************************************************/

/* упаковка 8-ми знаковых шортов (16S) в знаковые байты (8S) с насыщением
 * __di __builtin_e2k_packsshb (__di, __di) */

/* упаковка 8-ми знаковых шортов (16S) в беззнаковые байты (8U) с насыщением
 * __di __builtin_e2k_packushb (__di, __di) */

/* упаковка 4-х знаковых интов (32S) в знаковые шорты (16S) с насыщением
 * __di __builtin_e2k_packsswh (__di, __di) */

/* *****************************************************************************
 *                                Распаковка
 * *****************************************************************************/

/* чередование 4-х старших байт (8U) операнда 1 и 4-х старших байт (8U) операнда 2
 * __di __builtin_e2k_punpckhbh (__di, __di) */

/* чередование 4-х младших байт (8U) операнда 1 и 4-х младших байт (8U) операнда 2
 * __di __builtin_e2k_punpcklbh (__di, __di) */

/* чередование 2-х старших шортов (16U) операнда 1 и 2-х старших шортов (16U) операнда 2
 * __di __builtin_e2k_punpckhhw (__di, __di) */

/* чередование 2-х младших шортов (16U) операнда 1 и 2-х младших шортов (16U) операнда 2
 * __di __builtin_e2k_punpcklhw (__di, __di) */

/* чередование старшего инта (32U) операнда 1 и старшего инта (32U) операнда 2
 * __di __builtin_e2k_punpckhwd (__di, __di) */

/* чередование младшего инта (32U) операнда 1 и младшего инта (32U) операнда 2
 * __di __builtin_e2k_punpcklwd (__di, __di) */

/* *****************************************************************************
 *    Упакованные операции над числами c плавающей точкой одинарного формата
 * *****************************************************************************/

/* сложение 2-х флотов (32F)
 * __di __builtin_e2k_pfadds (__di, __di) */

/* вычитание 2-х флотов (32F)
 * __di __builtin_e2k_pfsubs (__di, __di) */

/* максимум 2-х флотов (32F)
 * __di __builtin_e2k_pfmaxs (__di, __di) */

/* минимум 2-х флотов (32F)
 * __di __builtin_e2k_pfmins (__di, __di) */

/* умножение 2-х флотов (32F)
 * __di __builtin_e2k_pfmuls (__di, __di) */

/* преобразование 2-х флотов (32F) в 2 инта (32S) в текущем режиме округления
 * __di __builtin_e2k_pfstois (__di) */

/* преобразование 2-х флотов (32F) в 2 инта (32S) c обрубанием
 * __di __builtin_e2k_pfstoistr (__di) */

/* преобразование 2-х интов (32S) в 2 флота (32F) в текущем режиме округления
 * __di __builtin_e2k_pistofs (__di) */

/* сравнение 2-х флотов (32F) на равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpeqs (__di, __di) */

/* сравнение 2-х флотов (32F) на меньше с формированием битовой маски
 * __di __builtin_e2k_pfcmplts (__di, __di) */

/* сравнение 2-х флотов (32F) на меньше или равно с формированием битовой маски
 * __di __builtin_e2k_pfcmples (__di, __di) */

/* сравнение 2-х флотов (32F) на неупорядочено с формированием битовой маски
 * __di __builtin_e2k_pfcmpuods (__di, __di) */

/* сравнение 2-х флотов (32F) на не равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpneqs (__di, __di) */

/* сравнение 2-х флотов (32F) на не меньше с формированием битовой маски
 * __di __builtin_e2k_pfcmpnlts (__di, __di) */

/* сравнение 2-х флотов (32F) на не меньше или равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpnles (__di, __di) */

/* сравнение 2-х флотов (32F) на упорядочено с формированием битовой маски
 * __di __builtin_e2k_pfcmpods (__di, __di) */

/* *****************************************************************************
 *      Операции преобразования из формата чисел с плавающей точкой в целые
 * *****************************************************************************/

/*  преобразование дабла (64F) в лонг (64S) в текущем режиме округления
 * __di __builtin_e2k_fdtoid (__di) */

/*  преобразование флота (32F) в инт (32S) в текущем режиме округления
 * unsigned __builtin_e2k_fstois (unsigned) */

/*  преобразование дабла (64F) в инт (32S) в текущем режиме округления
 * unsigned __builtin_e2k_fdtois (__di) */

/* *****************************************************************************
 *          Операции над числами c плавающей точкой одинарного формата
 * *****************************************************************************/

/* "упорядоченное" сравнение флотов (32F) с результатом в виде флагов
 * unsigned __builtin_e2k_fcmpodsf (unsigned, unsigned) */

/* "неупорядоченное" сравнение флотов (32F) с результатом в виде флагов
 * unsigned __builtin_e2k_fcmpudsf (unsigned, unsigned) */

/* *****************************************************************************
 *          Операции над числами c плавающей точкой двойного формата
 * *****************************************************************************/

/* максимум даблов (64F)
 * __di __builtin_e2k_pfmaxd (__di, __di) */

/* минимум даблов (64F)
 * __di __builtin_e2k_pfmind (__di, __di) */

/*  преобразование лонгa (64S) в дабл (64F) в текущем режиме округления
 * __di __builtin_e2k_idtofd (__di) */

/* сравнение даблов (64F) на равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpeqd (__di, __di) */

/* сравнение даблов (64F) на меньше с формированием битовой маски
 * __di __builtin_e2k_pfcmpltd (__di, __di) */

/* сравнение даблов (64F) на меньше или равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpled (__di, __di) */

/* сравнение даблов (64F) на неупорядочено с формированием битовой маски
 * __di __builtin_e2k_pfcmpuodd (__di, __di) */

/* сравнение даблов (64F) на не равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpneqd (__di, __di) */

/* сравнение даблов (64F) на не меньше с формированием битовой маски
 * __di __builtin_e2k_pfcmpnltd (__di, __di) */

/* сравнение даблов (64F) на не меньше или равно с формированием битовой маски
 * __di __builtin_e2k_pfcmpnled (__di, __di) */

/* сравнение даблов (64F) на упорядочено с формированием битовой маски
 * __di __builtin_e2k_pfcmpodd (__di, __di) */

/* "упорядоченное" сравнение даблов (64F) с результатом в виде флагов
 * unsigned __builtin_e2k_fcmpoddf (__di, __di) */

/* "неупорядоченное" сравнение даблов (64F) с результатом в виде флагов
 * unsigned __builtin_e2k_fcmpuddf (__di, __di) */

/* *****************************************************************************
 *                         Циклический битовый сдвиг
 * *****************************************************************************/

/* циклический сдвиг влево инта (32S/32U)
 * unsigned __builtin_e2k_scls (unsigned, unsigned) */

/* циклический сдвиг влево лонга (64S/64U)
 * __di __builtin_e2k_scld (__di, __di) */

/* циклический сдвиг вправо инта (32S/32U)
 * __di __builtin_e2k_scrs (unsigned, unsigned) */

/* циклический сдвиг вправо лонга (64S/64U)
 * __di __builtin_e2k_scrd (__di, __di) */

/* *****************************************************************************
 *         Сложение/вычитание с результатом в виде интелловских флагов
 * *****************************************************************************/

/* сложение интов (32S/32U)
 * unsigned __builtin_e2k_adds_fw (unsigned, unsigned) */

/* вычитание интов (32S/32U)
 * unsigned __builtin_e2k_subs_fw (unsigned, unsigned) */

/* *****************************************************************************
 *                              Вставление поля
 * *****************************************************************************/

/* вставление поля в инт (32S/32U)
 * unsigned __builtin_e2k_insfs (unsigned, unsigned, unsigned) */

/* вставление поля в лонг (64S/64U)
 * __di __builtin_e2k_insfd (__di, __di, __di) */

/* *****************************************************************************
 *                              Выделение поля
 * *****************************************************************************/

/* выделение поля из инта (32S/32U)
 * unsigned __builtin_e2k_getfs (unsigned, unsigned) */

/* выделение поля из лонга (64S/64U)
 * __di __builtin_e2k_getfd (__di, __di) */

/* *****************************************************************************
 *                        Вставить поле внешнего тэга
 * *****************************************************************************/

/* Осторожно: если компилятор вставит дополнительную безусловную пересылку
 * аргумента этой операции, то на ней может произойти падение еще до
 * выполнения операции */

/* вставление тэга в лонг (64S/64U) */
TWO_OPERANDS_INSTRUCTIONS (puttagd)

/* *****************************************************************************
 *                         Операции чтения/записи
 * *****************************************************************************/

/* чтение инта (32S/32U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_ld_32u_be(pval) __builtin_loadmas_32u (pval, 8, __LCC_CHAN_ANY)

/* чтение инта (32S/32U) без заведения в L1 кэше */
#define __builtin_e2k_ld_32u_t1(pval) __builtin_loadmas_32u (pval, 0x20, __LCC_CHAN_ANY)

/* чтение инта (32S/32U) без заведения в L1 и L2 кэше */
#define __builtin_e2k_ld_32u_t2(pval) __builtin_loadmas_32u (pval, 0x40, __LCC_CHAN_ANY)

/* чтение инта (32S/32U) без заведения во всех кэшах */
#define __builtin_e2k_ld_32u_nta(pval) __builtin_loadmas_32u (pval, 0x60, __LCC_CHAN_ANY)

/* чтение лонга (64S/64U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_ld_64s_be(pval) ((__di) __builtin_loadmas_64u (pval, 8, __LCC_CHAN_ANY))

/* чтение лонга (64S/64U) без заведения в L1 кэше */
#define __builtin_e2k_ld_64s_t1(pval) ((__di) __builtin_loadmas_64u (pval, 0x20, __LCC_CHAN_ANY))

/* чтение лонга (64S/64U) без заведения в L1 и L2 кэше */
#define __builtin_e2k_ld_64s_t2(pval) ((__di) __builtin_loadmas_64u (pval, 0x40, __LCC_CHAN_ANY))

/* чтение лонга (64S/64U) без заведения во всех кэшах */
#define __builtin_e2k_ld_64s_nta(pval) ((__di) __builtin_loadmas_64u (pval, 0x60, __LCC_CHAN_ANY))

/* чтение лонга (64S/64U) в спекулятивном режиме */
#define __builtin_e2k_ld_64s_sm(pval, ind) ((__di) __builtin_loadmas_64u ((char *) (pval) + (ind), __LCC_MAS_SPEC, __LCC_CHAN_ANY))

/* чтение лонга (64S/64U) в спекулятивном режиме c обнулением тега */
#define __builtin_e2k_ld_64s_cleartag(pval, ind) __builtin_loadmas_64u ((char *) (pval) + (ind), __LCC_MAS_SPEC | __LCC_MAS_CLEARTAG, __LCC_CHAN_ANY)

/* запись инта (32S/32U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_st_32u_be(val, pval) __builtin_storemas_32u (val, pval, 8, __LCC_CHAN_ANY)

/* запись инта (32S/32U) мимо кэшей, напрямую в память (non temporal) */
#define __builtin_e2k_st_32u_nt(val, pval) __builtin_storemas_32u (val, pval, 0x60, __LCC_CHAN_ANY)

/* запись лонга (64S/64U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_st_64s_be(val, pval) __builtin_storemas_64u (val, pval, 8, __LCC_CHAN_ANY)

/* запись лонга (64S/64U) мимо кэшей, напрямую в память (non temporal) */
#define __builtin_e2k_st_64s_nt(val, pval) __builtin_storemas_64u (val, pval, 0x60, __LCC_CHAN_ANY)

/* Очистка строки кэша и удаление ее из всех кэшей */
#define __builtin_e2k_clflush(p) __builtin_storemas_64u (0, (void *) (p), 0xf, 2)

/* константы для использования с prefetch */
enum __e2k_hint {
    E2K_HINT_T0 = 3,  /* обычное функционирование кэш-памятей всех уровней */
    E2K_HINT_T1 = 2,  /* запрещение заведения строки в DCACHE1 */
    E2K_HINT_T2 = 1,  /* запрещение заведения строки в DCACHE1, DCACHE2 */
    E2K_HINT_NTA = 0  /* запрещение заведения строки в DCACHE1, DCACHE2, ECACHE */
};

/* подкачка строки в кэш соответствующего уровня */
#define __builtin_e2k_prefetch(P, I) \
    __builtin_prefetch (/* void * */(P), 0, /* __e2k_hint */(I))


#if __iset__ >= 2 /* __elbrus_2cplus__ и старше */

/* *****************************************************************************
 *                         Новые операции в Эльбрус-2C+
 * *****************************************************************************
 *                   Перестановка разрядов в обратном порядке
 * *****************************************************************************/

/* перестановка разрядов в обратном порядке в инте (32S/32U)
 * unsigned __builtin_e2k_bitrevs (unsigned) */

/* перестановка разрядов в обратном порядке в лонге (64S/64U)
 * __di __builtin_e2k_bitrevd (__di) */

/* *****************************************************************************
 *                      Подсчет количества лидирующих нулей
 * *****************************************************************************/

/* подсчет количества лидирующих нулей в инте (32S/32U)
 * unsigned __builtin_e2k_lzcnts (unsigned) */

/* подсчет количества лидирующих нулей в лонге (64S/64U)
 * __di __builtin_e2k_lzcntd (__di) */

/* *****************************************************************************
 *                         Подсчет количества единиц
 * *****************************************************************************/

/* подсчет количества единиц в инте (32S/32U)
 * unsigned __builtin_e2k_popcnts (unsigned) */

/* подсчет количества единиц в лонге (64S/64U)
 * __di __builtin_e2k_popcntd (__di) */

/* *****************************************************************************
 *                          Умножение байт на шорты
 * *****************************************************************************/

/* умножение 4-х беззнаковых байт (8U) на 4-ре знаковых шорта (16S) со сдвигом
 * результата вправо на 8 и округлением к ближайшему
 * __di __builtin_e2k_pmulubhh (__di, __di) */

/* *****************************************************************************
 *                             Перестановка байт
 * *****************************************************************************/

/* выделение из 16-ти входных байт восьми, положение которых определяется значениями
 * в соответствующих байтах 3-го операнда. Либо заполнение предопределенными константами.
 * Для v4 и далее - дополнительно работает инверсия, реверсивная перестановка бит
 * или заполнение знаком
 * __di __builtin_e2k_pshufb (__di, __di, __di) */

/* *****************************************************************************
 *                                Выбор байт
 * *****************************************************************************/

/* выбор 8-ми байт либо из операнда 1, либо из операнда 2 в зависимости от значения масок,
 * содержащихся в операнде 3
 * __di __builtin_e2k_pmerge (__di, __di, __di) */

/* *****************************************************************************
 *          Операции над числами c плавающей точкой двойного формата
 * *****************************************************************************/

/*  преобразование дабла (64F) в лонг (64S) с обрубанием
 * __di __builtin_e2k_fdtoidtr (__di) */

#endif /* __iset__ >= 2 */


#if __iset__ >= 3 /* __elbrus_4c__ и старше */

/* *****************************************************************************
 *                         Новые операции в Эльбрус-4C
 * *****************************************************************************/

/* Операции psrl(w/h), psra(w/h), psll(w/h) добавлены в 0/3 каналы */

/* *****************************************************************************
 *                  Горизонтальное сложение и вычитание по модулю
 * *****************************************************************************/

/* Аргументы выкладываются в обратном порядке перед операцией : src2, src1 */

/* четыре горизонтальных сложения соседних шортов (16S/16U)
 * __di __builtin_e2k_phaddh (__di, __di) */

/* два горизонтальных сложения соседних интов (32S/32U)
 * __di __builtin_e2k_phaddw (__di, __di) */

/* четыре горизонтальных вычитания соседних шортов (16S/16U)
 * __di __builtin_e2k_phsubh (__di, __di) */

/* два горизонтальных вычитания соседних интов (32S/32U)
 * __di __builtin_e2k_phsubw (__di, __di) */

/* *****************************************************************************
 *                Горизонтальное сложение и вычитание c насыщением
 * *****************************************************************************/

/* четыре горизонтальных сложения соседних знаковых шортов (16S)
 * __di __builtin_e2k_phaddsh (__di, __di) */

/* четыре горизонтальных вычитания соседних знаковых шортов (16S)
 * __di __builtin_e2k_phsubsh (__di, __di) */

/* *****************************************************************************
 *                             Максимум и минимум
 * *****************************************************************************/

/* максимум 8-ми знаковых байт (8S)
 * __di __builtin_e2k_pmaxsb (__di, __di) */

/* максимум 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_pmaxuh (__di, __di) */

/* максимум 2-х беззнаковых интов (32U)
 * __di __builtin_e2k_pmaxuw (__di, __di) */

/* максимум 2-х знаковых интов (32S)
 * __di __builtin_e2k_pmaxsw (__di, __di) */

/* минимум 8-ми знаковых байт (8S)
 * __di __builtin_e2k_pminsb (__di, __di) */

/* минимум 4-х беззнаковых шортов (16U)
 * __di __builtin_e2k_pminuh (__di, __di) */

/* минимум 2-х беззнаковых интов (32U)
 * __di __builtin_e2k_pminuw (__di, __di) */

/* минимум 2-х знаковых интов (32S)
 * __di __builtin_e2k_pminsw (__di, __di) */

/* *****************************************************************************
 *                             Минимум и его позиция
 * *****************************************************************************/

/* поиск минимального из 8-ми беззнаковых шортов (16U) и его позиции
 * __di __builtin_e2k_phminposuh (__di, __di) */

/* *****************************************************************************
 *                 Умножение байт с горизонтальным сложением
 * *****************************************************************************/

/* умножение 8-ми знаковых байт (8S) на 8 беззнаковых байт (8U) и попарное сложение
 * смежных шортовых произведений с сатурацией
 * __di __builtin_e2k_pmaddubsh (__di, __di) */

/* *****************************************************************************
 *                    Умножение шортов с масштабированием
 * *****************************************************************************/

/* умножение 4-х знаковых шортов (16S) c масштабированием (>> 15) и округлением
 * __di __builtin_e2k_pmulhrsh (__di, __di) */

/* *****************************************************************************
 *                             Умножение лонгов
 * *****************************************************************************/

/* умножение беззнаковых лонгов (64S) и выдача старших 64 разрядов произведения
 * __di __builtin_e2k_umulhd (__di, __di) */

/* умножение знаковых лонгов (64S) и выдача старших 64 разрядов произведения
 * __di __builtin_e2k_smulhd (__di, __di) */

/* *****************************************************************************
 *         Сложение/вычитание с результатом в виде интелловских флагов
 * *****************************************************************************/

/* сложение лонгов (64S/64U)
 * unsigned __builtin_e2k_addd_fd (__di, __di) */

/* вычитание лонгов (64S/64U)
 * unsigned __builtin_e2k_subd_fd (__di, __di) */

/* *****************************************************************************
 *         Множественное сложение абсолютных разностей беззнаковых байт
 * *****************************************************************************/

/* Четыре cложения абсолютных разностей 4-х беззнаковых байт (8U) со сдвигом операндов
 * первого аргумента и 4-х беззнаковых байт (8U) второго аргумента с формированием 4-х
 * беззнаковых шортов (16U)
 * __di __builtin_e2k_mpsadbh (__di, __di) */

/* *****************************************************************************
 *                             Cравнение лонгов
 * *****************************************************************************/

/* сравнение на "равно" лонгов (64S/64U)
 * __di __builtin_e2k_pcmpeqd (__di, __di) */

/* сравнение на "больше" знаковых лонгов (64S)
 * __di __builtin_e2k_pcmpgtd (__di, __di) */

/* *****************************************************************************
 *                               Упаковка интов
 * *****************************************************************************/

/* упаковка 4-х знаковых интов (32S) в беззнаковые шорты (16U) с насыщением
 * __di __builtin_e2k_packuswh (__di, __di) */

/* *****************************************************************************
 *                            Изменение знака
 * *****************************************************************************/

/* умножение 8-ми знаковых байт (8S) на знаки 8-ми байт (8S) второго аргумента
 * __di __builtin_e2k_psignb (__di, __di) */

/* умножение 4-х знаковых шортов (16S) на знаки 4-х шортов (16S) второго аргумента
 * __di __builtin_e2k_psignh (__di, __di) */

/* умножение 2-х знаковых интов (32S) на знаки 2-х интов (32S) второго аргумента
 * __di __builtin_e2k_psignw (__di, __di) */

/* *****************************************************************************
 *    Упакованные операции над числами c плавающей точкой одинарного формата
 * *****************************************************************************/

/* два горизонтальных сложения соседних флотов (32F)
 * __di __builtin_e2k_pfhadds (__di, __di) */

/* два горизонтальных вычитания соседних флотов (32F)
 * __di __builtin_e2k_pfhsubs (__di, __di) */

/* сложение старших и вычитание младших 2-х флотов (32F)
 * __di __builtin_e2k_pfaddsubs (__di, __di) */

/* *****************************************************************************
 *                 Операции преобразования c плавающей точкой
 * *****************************************************************************/

/* управление округлением - первый аргумент 3-х следующих операций */
#define _TOIF_RC_NEAREST        0x0
#define _TOIF_RC_DOWN           0x1
#define _TOIF_RC_UP             0x2
#define _TOIF_RC_ZERO           0x3
#define _TOIF_RC_CURRENT        0x4 /* режим округления из регистра pfpfr */
#define _TOIF_RC_IGNORE_INEXACT 0x8 /* исключение inexact не вырабатывается */

/* выделение целой части значения флота (32F) в заданном режиме округления
 * unsigned __builtin_e2k_fstoifs (unsigned, unsigned) */

/* выделение целой части значений 2-х флотов (32F) в заданном режиме округления
 * __di __builtin_e2k_pfstoifs (__di, __di) */

/*  выделение целой части значения дабла (64F) в заданном режиме округления
 * __di __builtin_e2k_fdtoifd (__di, __di) */

#endif /*  __iset__ >= 3 */

#if __iset__ >= 4 /* __elbrus_8c__ и __elbrus_1cplus__ и старше */

/* *****************************************************************************
 *                      Новые операции в Эльбрус-1C+/8C
 * *****************************************************************************
 *              Умножение вещественного числа на целую степень двойки
 * *****************************************************************************/

/* умножение флота (32F) на целую степень двойки
 * unsigned __builtin_e2k_fscales (unsigned, sint32) */

/* умножение дабла (64F) на целую степень двойки
 * __di __builtin_e2k_fscaled (__di, sint32) */

/* умножение лонг дабла (80F) на целую степень двойки
long double __builtin_e2k_fxscalesx (long double src1, sint32 src2) */

/* расширены возможности операции pshufb - теперь она может дополнительно делать
 * инвертирование байта (4 бит), обратную перестановку бит (5 бит) и размножение
 * знака (6 бит) */

#endif /* __iset__ >= 4 */

#if __iset__ >= 5 /* __elbrus_8c2__ и старше */

/* *****************************************************************************
 *                         Новые операции в Эльбрус-8C2
 * *****************************************************************************
 *                         Сложение и вычитание по модулю
 * *****************************************************************************/

/* сложение 16-ти байт (8S/8U)
 * __v2di __builtin_e2k_qpaddb (__v2di, __v2di) */

/* сложение 8-ми шортов (16S/16U)
 * __v2di __builtin_e2k_qpaddh (__v2di, __v2di) */

/* сложение 4-х интов (32S/32U)
 * __v2di __builtin_e2k_qpaddw (__v2di, __v2di) */

/* сложение 2-х лонгов (64S/64U)
 * __v2di __builtin_e2k_qpaddd (__v2di, __v2di) */

/* вычитание 16-ти байт (8S/8U)
 * __v2di __builtin_e2k_qpsubb (__v2di, __v2di) */

/* вычитание 8-ми шортов (16S/16U)
 * __v2di __builtin_e2k_qpsubh (__v2di, __v2di) */

/* вычитание 4-х интов (32S/32U)
 * __v2di __builtin_e2k_qpsubw (__v2di, __v2di) */

/* вычитание 2-х лонгов (64S/64U)
 * __v2di __builtin_e2k_qpsubd (__v2di, __v2di) */

/* *****************************************************************************
 *                  Горизонтальное сложение и вычитание по модулю
 * *****************************************************************************/

/* восемь горизонтальных сложений соседних шортов (16S/16U)
 * __v2di __builtin_e2k_qphaddh (__v2di, __v2di) */

/* четыре горизонтальных сложения соседних интов (32S/32U)
 * __v2di __builtin_e2k_qphaddw (__v2di, __v2di) */

/* восемь горизонтальных вычитаний соседних шортов (16S/16U)
 * __v2di __builtin_e2k_qphsubh (__v2di, __v2di) */

/* четыре горизонтальных вычитания соседних интов (32S/32U)
 * __v2di __builtin_e2k_qphsubw (__v2di, __v2di) */

/* *****************************************************************************
 *                         Сложение и вычитание c насыщением
 * *****************************************************************************/

/* сложение 16-ти знаковых байт (8S)
 * __v2di __builtin_e2k_qpaddsb (__v2di, __v2di) */

/* сложение 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpaddsh (__v2di, __v2di) */

/* сложение 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpaddusb (__v2di, __v2di) */

/* сложение 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpaddush (__v2di, __v2di) */

/* вычитание 16-ти знаковых байт (8S)
 * __v2di __builtin_e2k_qpsubsb (__v2di, __v2di) */

/* вычитание 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpsubsh (__v2di, __v2di) */

/* вычитание 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpsubusb (__v2di, __v2di) */

/* вычитание 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpsubush (__v2di, __v2di) */

/* *****************************************************************************
 *                Горизонтальное сложение и вычитание c насыщением
 * *****************************************************************************/

/* восемь горизонтальных сложений соседних знаковых шортов (16S)
 * __v2di __builtin_e2k_qphaddsh (__v2di, __v2di) */

/* восемь горизонтальных вычитаний соседних знаковых шортов (16S)
 * __v2di __builtin_e2k_qphsubsh (__v2di, __v2di) */

/* *****************************************************************************
 *                             Максимум и минимум
 * *****************************************************************************/

/* максимум 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpmaxub (__v2di, __v2di) */

/* максимум 16-ти знаковых байт (8S)
 * __v2di __builtin_e2k_qpmaxsb (__v2di, __v2di) */

/* максимум 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpmaxuh (__v2di, __v2di) */

/* максимум 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpmaxsh (__v2di, __v2di) */

/* максимум 4-х беззнаковых интов (32U)
 * __v2di __builtin_e2k_qpmaxuw (__v2di, __v2di) */

/* максимум 4-х знаковых интов (32S)
 * __v2di __builtin_e2k_qpmaxsw (__v2di, __v2di) */

/* минимум 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpminub (__v2di, __v2di) */

/* минимум 16-ти знаковых байт (8S)
 * __v2di __builtin_e2k_qpminsb (__v2di, __v2di) */

/* минимум 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpminuh (__v2di, __v2di) */

/* минимум 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpminsh (__v2di, __v2di) */

/* минимум 4-х беззнаковых интов (32U)
 * __v2di __builtin_e2k_qpminuw (__v2di, __v2di) */

/* минимум 4-х знаковых интов (32S)
 * __v2di __builtin_e2k_qpminsw (__v2di, __v2di) */

/* *****************************************************************************
 *                             Минимум и его позиция
 * *****************************************************************************/

/* поиск минимального из 16-ти беззнаковых шортов (16U) и его позиции
 * __di __builtin_e2k_qphminposuh (__v2di, __v2di) */

/* *****************************************************************************
 *                             Умножение шортов
 * *****************************************************************************/

/* умножение 8-ми беззнаковых шортов (16U) и выдача старших 16 разрядов произведений
 * __v2di __builtin_e2k_qpmulhuh (__v2di, __v2di) */

/* умножение 8-ми знаковых шортов (16S) и выдача старших 16 разрядов произведений
 * __v2di __builtin_e2k_qpmulhh (__v2di, __v2di) */

/* умножение 8-ми шортов (16S/16U) и выдача младших 16 разрядов произведений
 * __v2di __builtin_e2k_qpmullh (__v2di, __v2di) */

/* *****************************************************************************
 *                             Умножение интов
 * *****************************************************************************/

/* умножение 2-х интов (32S/32U) и выдача младших 32 разрядов произведений
 * __di __builtin_e2k_pmullw (__di, __di) */

/* умножение 4-х интов (32S/32U) и выдача младших 32 разрядов произведений
 * __v2di __builtin_e2k_qpmullw (__v2di, __v2di) */

/* *****************************************************************************
 *                          Умножение байт на шорты
 * *****************************************************************************/

/* умножение 8-ми беззнаковых байт (8U) на 8-мь знаковых шортов (16S) со сдвигом
 * результата вправо на 8 и округлением к ближайшему
 * __v2di __builtin_e2k_qpmulubhh (__di, __v2di) */

/* *****************************************************************************
 *                    Умножение шортов с масштабированием
 * *****************************************************************************/

/* умножение 8-ми знаковых шортов (16S) c масштабированием (>> 15) и округлением
 * __v2di __builtin_e2k_qpmulhrsh (__v2di, __v2di) */

/* *****************************************************************************
 *                 Умножение шортов с горизонтальным сложением
 * *****************************************************************************/

/* умножение 8-ми знаковых шортов (16S) и попарное сложение смежных интовых произведений
 * __v2di __builtin_e2k_qpmaddh (__v2di, __v2di) */

/* *****************************************************************************
 *                 Умножение байт с горизонтальным сложением
 * *****************************************************************************/

/* умножение 16-ти знаковых байт (8S) на 16 беззнаковых байт (8U) и попарное сложение
 * смежных шортовых произведений с сатурацией
 * __v2di __builtin_e2k_qpmaddubsh (__v2di, __v2di) */

/* *****************************************************************************
 *                            Изменение знака
 * *****************************************************************************/

/* умножение 16-ти знаковых байт (8S) на знаки 16-ти байт (8S) второго аргумента
 * __v2di __builtin_e2k_qpsignb (__v2di, __v2di) */

/* умножение 8-ми знаковых шортов (16S) на знаки 8-ми шортов (16S) второго аргумента
 * __v2di __builtin_e2k_qpsignh (__v2di, __v2di) */

/* умножение 4-х знаковых интов (32S) на знаки 4-х интов (32S) второго аргумента
 * __v2di __builtin_e2k_qpsignw (__v2di, __v2di) */

/* *****************************************************************************
 *                Cложение абсолютных разностей беззнаковых байт
 * *****************************************************************************/

/* cложение абсолютных разностей 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpsadbw (__v2di, __v2di) */

/* *****************************************************************************
 *         Множественное сложение абсолютных разностей беззнаковых байт
 * *****************************************************************************/

/* две операции mpsadbh над младшей и старшей половинами операнда 1 и общим
 * операндом 2
 * __v2di __builtin_e2k_qpmpsadbh (__v2di, unsigned) */

/* *****************************************************************************
 *                         Среднее арифметическое
 * *****************************************************************************/

/* среднее 16-ти беззнаковых байт (8U)
 * __v2di __builtin_e2k_qpavgusb (__v2di, __v2di) */

/* среднее 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpavgush (__v2di, __v2di) */

/* *****************************************************************************
 *                                Cравнение
 * *****************************************************************************/

/* сравнение на "равно" 16-ти байт (8S/8U)
 * __v2di __builtin_e2k_qpcmpeqb (__v2di, __v2di) */

/* сравнение на "равно" 8-ми шортов (16S/16U)
 * __v2di __builtin_e2k_qpcmpeqh (__v2di, __v2di) */

/* сравнение на "равно" 4-х интов (32S/32U)
 * __v2di __builtin_e2k_qpcmpeqw (__v2di, __v2di) */

/* сравнение на "равно" 2-х лонгов (64S/64U)
 * __v2di __builtin_e2k_qpcmpeqd (__v2di, __v2di) */

/* сравнение на "больше" 16-ти знаковых байт (8S)
 * __v2di __builtin_e2k_qpcmpgtb (__v2di, __v2di) */

/* сравнение на "больше" 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpcmpgth (__v2di, __v2di) */

/* сравнение на "больше" 4-х знаковых интов (32S)
 * __v2di __builtin_e2k_qpcmpgtw (__v2di, __v2di) */

/* сравнение на "больше" 2-х знаковых лонгов (64S)
 * __v2di __builtin_e2k_qpcmpgtd (__v2di, __v2di) */

/* *****************************************************************************
 *                             Перестановка байт
 * *****************************************************************************/

/* выделение из 32-х входных байт 16-ти, положение которых определяется значениями
 * в соответствующих байтах 3-го операнда. Выделение байт возможно только локально
 * внутри старшей и младшей половин аргументов и результата. Далее возможно
 * заполнение предопределенными константами, инверсия, реверсивная перестановка бит
 * или заполнение знаком. Поле [3:0] определяет номер байта, а поле [7:4] - тип
 * операции над выбранным байтом
 * __v2di __builtin_e2k_qpshufb (__v2di, __v2di, __v2di) */

/* выделение из 32-х входных байт произвольных 16-ти, положение которых определяется
 * значениями в соответствующих байтах 3-го операнда и выполнение над ними логических
 * операций в зависимости от значений управляющих байт в src3. Поле [4:0] определяет
 * номер байта, а поле [7:5] - тип операции над выбранным байтом
 * __v2di __builtin_e2k_qppermb (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                                Выбор байт
 * *****************************************************************************/

/* выбор 16-ти байт либо из операнда 1, либо из операнда 2 в зависимости от значения
 * масок, содержащихся в операнде 3
 * __v2di __builtin_e2k_qpmerge (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                                Упаковка лонгов
 * *****************************************************************************/

/* упаковка 2-х лонгов (64S/64U) в квадро-регистр
 * __v2di __builtin_e2k_qppackdl (__di, __di) */

/* *****************************************************************************
 *                                Перестановки
 * *****************************************************************************/

/* меняет местами двойные слова (64S)
 * __v2di __builtin_e2k_qpswitchd (__v2di) */

/* меняет местами значения внутри 2-х пар интов (32S)
 * __v2di __builtin_e2k_qpswitchw (__v2di) */

/* *****************************************************************************
 *                             Формирование маски
 * *****************************************************************************/

/* формирование маски из знака байтов
 * unsigned __builtin_e2k_qpsgn2mskb (__v2di) */

/* *****************************************************************************
 *                          Модификация разрядов знака
 * *****************************************************************************/

/* модификация разрядов знака упакованных байтов (8S) значениями битов 16-разрядной
 * инверсной маски
 * __v2di __builtin_e2k_qpmsk2sgnb (__v2di, unsigned) */

/* *****************************************************************************
 *                           Логические операции
 * *****************************************************************************/

/* поразрядное логическое AND 128-и разрядов (128U)
 * __v2di __builtin_e2k_qpand (__v2di, __v2di) */

/* поразрядное логическое AND 128-и разрядов c инверсией второго операнда (128U)
 * __v2di __builtin_e2k_qpandn (__v2di, __v2di) */

/* поразрядное логическое OR 128-и разрядов (128U)
 * __v2di __builtin_e2k_qpor (__v2di, __v2di) */

/* поразрядное логическое XOR 128-и разрядов (128U)
 * __v2di __builtin_e2k_qpxor (__v2di, __v2di) */

/* *****************************************************************************
 *                       Произвольные логическая операции
 * *****************************************************************************/

/* произвольная поразрядная логическая операция над 64 разрядами (64S/64U)
 * __di __builtin_e2k_plog (N, __di, __di, __di) */

/* произвольная поразрядная логическая операция над 128 разрядами (128U)
 * __v2di __builtin_e2k_qplog (N, __v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                               Битовый сдвиг
 * *****************************************************************************/

/* сдвиг влево 8-ми шортов (16S/16U)
 * __v2di __builtin_e2k_qpsllh (__v2di, __di) */

/* сдвиг влево 4-х интов (32S/32U)
 * __v2di __builtin_e2k_qpsllw (__v2di, __di) */

/* сдвиг влево 2-х лонгов (64S/64U)
 * __v2di __builtin_e2k_qpslld (__v2di, __di) */

/* логический сдвиг вправо 8-ми беззнаковых шортов (16U)
 * __v2di __builtin_e2k_qpsrlh (__v2di, __di) */

/* логический сдвиг вправо 4-х беззнаковых интов (32U)
 * __v2di __builtin_e2k_qpsrlw (__v2di, __di) */

/* логический сдвиг вправо 2-х беззнаковых лонгов (64U)
 * __v2di __builtin_e2k_qpsrld (__v2di, __di) */

/* арифметический сдвиг вправо 8-ми знаковых шортов (16S)
 * __v2di __builtin_e2k_qpsrah (__v2di, __di) */

/* арифметический сдвиг вправо 4-х знаковых интов (32S)
 * __v2di __builtin_e2k_qpsraw (__v2di, __di) */

/* *****************************************************************************
 *                       Циклический битовый сдвиг вправо
 * *****************************************************************************/

/* циклический сдвиг вправо 2-х интов (32S/32U)
 * __di __builtin_e2k_psrcw (__di, __di) */

/* циклический сдвиг вправо лонга (64S/64U)
 * __di __builtin_e2k_psrcd (__di, __di) */

/* циклический сдвиг вправо 4-х интов (32S/32U)
 * __v2di __builtin_e2k_qpsrcw (__v2di, __di) */

/* циклический сдвиг вправо 2-х лонгов (64S/64U)
 * __v2di __builtin_e2k_qpsrcd (__v2di, __di) */

/* *****************************************************************************
 *                                 Упаковка
 * *****************************************************************************/

/* упаковка 16-ти знаковых шортов (16S) в знаковые байты (8S) с насыщением
 * __v2di __builtin_e2k_qpacksshb (__v2di, __v2di) */

/* упаковка 16-ти знаковых шортов (16S) в беззнаковые байты (8U) с насыщением
 * __v2di __builtin_e2k_qpackushb (__v2di, __v2di) */

/* упаковка 8-ми знаковых интов (32S) в знаковые шорты (16S) с насыщением
 * __v2di __builtin_e2k_qpacksswh (__v2di, __v2di) */

/* упаковка 8-ми знаковых интов (32S) в беззнаковые шорты (16U) с насыщением
 * __v2di __builtin_e2k_qpackuswh (__v2di, __v2di) */

/* *****************************************************************************
 *                                Распаковка
 * *****************************************************************************/

/* чередование 8-ми старших байт (8U) операнда 1 и 8-ми старших байт (8U) операнда 2
 * __v2di __builtin_e2k_qpunpckhbh (__v2di, __v2di) */

#define __builtin_e2k_qpunpckhbh(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1f0f1e0e1d0d1c0cLL, 0x1b0b1a0a19091808LL))

/* чередование 8-ми младших байт (8U) операнда 1 и 8-ми младших байт (8U) операнда 2
 * __v2di __builtin_e2k_qpunpcklbh (__v2di, __v2di) */

#define __builtin_e2k_qpunpcklbh(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1707160615051404LL, 0x1303120211011000LL))

/* чередование 4-х старших шортов (16U) операнда 1 и 4-х старших шортов (16U) операнда 2
 * __v2di __builtin_e2k_qpunpckhhw (__v2di, __v2di) */

#define __builtin_e2k_qpunpckhhw(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1f1e0f0e1d1c0d0cLL, 0x1b1a0b0a19180908LL))

/* чередование 4-х младших шортов (16U) операнда 1 и 4-х младших шортов (16U) операнда 2
 * __v2di __builtin_e2k_qpunpcklhw (__v2di, __v2di) */

#define __builtin_e2k_qpunpcklhw(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1716070615140504LL, 0x1312030211100100LL))

/* чередование 2-х старших интов (32U) операнда 1 и 2-х старших интов (32U) операнда 2
 * __v2di __builtin_e2k_qpunpckhwd (__v2di, __v2di) */

#define __builtin_e2k_qpunpckhwd(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1f1e1d1c0f0e0d0cLL, 0x1b1a19180b0a0908LL))

/* чередование 2-х младших интов (32U) операнда 1 и 2-х младших интов (32U) операнда 2
 * __v2di __builtin_e2k_qpunpcklwd (__v2di, __v2di) */

#define __builtin_e2k_qpunpcklwd(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1716151407060504LL, 0x1312111003020100LL))

/* чередование старшего лонга (64U) операнда 1 и старшего лонга (64U) операнда 2
 * __v2di __builtin_e2k_qpunpckhdq (__v2di, __v2di) */

#define __builtin_e2k_qpunpckhdq(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1f1e1d1c1b1a1918LL, 0x0f0e0d0c0b0a0908LL))

/* чередование младшего лонга (64U) операнда 1 и младшего лонга (64U) операнда 2
 * __v2di __builtin_e2k_qpunpckldq (__v2di, __v2di) */

#define __builtin_e2k_qpunpckldq(src1, src2)                                                  \
    __builtin_e2k_qppermb (src1, src2, __builtin_e2k_qppackdl (0x1716151413121110LL, 0x0706050403020100LL))

/* *****************************************************************************
 *                Выделение поля c обнулением младших разрядов
 * *****************************************************************************/

/* выделение поля из инта (32S/32U) c обнулением младших разрядов
 * unsigned __builtin_e2k_getfzs (unsigned, unsigned) */

/* выделение поля из лонга (64S/64U) c обнулением младших разрядов
 * __di __builtin_e2k_getfzd (__di, __di) */

/* *****************************************************************************
 *                        Вставить поле внешнего тэга
 * *****************************************************************************/

/* Осторожно: если компилятор вставит дополнительную безусловную пересылку
 * аргумента этой операции, то на ней может произойти падение еще до
 * выполнения операции */

/* вставление тэга в 128-ми разрядное значение (128U) */
TWO_OPERANDS_INSTRUCTIONS_QSQ (puttagqp)

/* *****************************************************************************
 *             Сложение и вычитание c битом переноса (carry)
 * *****************************************************************************/

/* сложение лонгов (64S/64U) и бита переноса c выработкой числового результата
 * __di __builtin_e2k_addcd (__di, __di, unsigned) */

/* вычитание лонгов (64S/64U) и бита переноса c выработкой числового результата
 * __di __builtin_e2k_subcd (__di, __di, unsigned) */

/* сложение лонгов (64S/64U) и бита переноса c выработкой нового переноса
 * __di __builtin_e2k_addcd_c (__di, __di, unsigned) */

/* вычитание лонгов (64S/64U) и бита переноса c выработкой нового переноса
 * __di __builtin_e2k_subcd_c (__di, __di, unsigned) */

/* *****************************************************************************
 *    Упакованные операции над числами c плавающей точкой одинарного формата
 * *****************************************************************************/

/* сложение 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfadds (__v2di, __v2di) */

/* горизонтальное сложение 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfhadds (__v2di, __v2di) */

/* вычитание 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfsubs (__v2di, __v2di) */

/* горизонтальное вычитание 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfhsubs (__v2di, __v2di) */

/* сложение нечетных и вычитание четных 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfaddsubs (__v2di, __v2di) */

/* максимум 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfmaxs (__v2di, __v2di) */

/* минимум 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfmins (__v2di, __v2di) */

/* умножение 4-х флотов (32F)
 * __v2di __builtin_e2k_qpfmuls (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpeqs (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на меньше с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmplts (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на меньше или равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmples (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на не упорядочено с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpuods (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на не равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpneqs (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на не меньше с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpnlts (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на не меньше или равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpnles (__v2di, __v2di) */

/* сравнение 4-х флотов (32F) на упорядочено с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpods (__v2di, __v2di) */

/* преобразование 4-х флотов (32F) в 4 инта (32S) в текущем режиме округления
 * __v2di __builtin_e2k_qpfstois (__v2di) */

/* преобразование 4-х флотов (32F) в 4 инта (32S) c обрубанием
 * __v2di __builtin_e2k_qpfstoistr (__v2di) */

/* преобразование 4-х интов (32S) в 4 флота (32F) в текущем режиме округления
 * __v2di __builtin_e2k_qpistofs (__v2di) */

/* преобразование 2-х лонгов (64S) в 2 флота (32F)
 * __di __builtin_e2k_qpidtofs (__v2di) */

/* выделение целой части значений 4-х флотов (32F) в заданном режиме округления
 * __v2di __builtin_e2k_qpfstoifs (__di, __v2di) */

/* преобразование 2-х флотов (32F) в 2 дабла (64F)
 * __v2di __builtin_e2k_qpfstofd (__di) */

/* преобразование 2-х флотов (32F) в 2 лонга (64S) в текущем режиме округления
 * __v2di __builtin_e2k_qpfstoid (__di) */

/* преобразование 2-х флотов (32F) в 2 лонга (64S) c обрубанием
 * __v2di __builtin_e2k_qpfstoidtr (__di) */

/* *****************************************************************************
 *     Упакованные операции над числами c плавающей точкой двойного формата
 * *****************************************************************************/

/* сложение 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfaddd (__v2di, __v2di) */

/* вычитание 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfsubd (__v2di, __v2di) */

/* сложение старших и вычитание младших 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfaddsubd (__v2di, __v2di) */

/* максимум 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfmaxd (__v2di, __v2di) */

/* минимум 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfmind (__v2di, __v2di) */

/* умножение 2-х даблов (64F)
 * __v2di __builtin_e2k_qpfmuld (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpeqd (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на меньше с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpltd (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на меньше или равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpled (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на не упорядочено с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpuodd (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на не равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpneqd (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на не меньше с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpnltd (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на не меньше или равно с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpnled (__v2di, __v2di) */

/* сравнение 2-х даблов (64F) на упорядочено с формированием битовой маски
 * __v2di __builtin_e2k_qpfcmpodd (__v2di, __v2di) */

/* преобразование 2-х даблов (64F) в 2 инта (32S) в текущем режиме округления
 * __di __builtin_e2k_qpfdtois (__v2di) */

/* преобразование 2-х даблов (64F) в 2 инта (32S) c обрубанием
 * __di __builtin_e2k_qpfdtoistr (__v2di) */

/* преобразование 2-х интов (32S) в 2 дабла (64F)
 * __v2di __builtin_e2k_qpistofd (__di) */

/* преобразование 2-х лонгов (64S) в 2 дабла (64F)
 * __v2di __builtin_e2k_qpidtofd (__v2di) */

/* выделение целой части значений 2-х даблов (64F) в заданном режиме округления
 * __v2di __builtin_e2k_qpfdtoifd (__di, __v2di) */

/* преобразование 2-х даблов (64F) в 2 флота (32F)
 * __di __builtin_e2k_qpfdtofs (__v2di) */

/* преобразование 2-х даблов (64F) в 2 лонга (64S) в текущем режиме округления
 * __v2di __builtin_e2k_qpfdtoid (__v2di) */

/* преобразование 2-х даблов (64F) в 2 лонга (64S) c обрубанием
 * __v2di __builtin_e2k_qpfdtoidtr (__v2di) */

/* *****************************************************************************
 *                         Операции чтения/записи
 * *****************************************************************************/

/* чтение 16 байт (2*64U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_ld_128_be(pval) __builtin_loadmas_128v (pval, 8, __LCC_CHAN_ANY)

/* чтение 16 байт (2*64U) без заведения в L1 кэше */
#define __builtin_e2k_ld_128_t1(pval) __builtin_loadmas_128v (pval, 0x20, __LCC_CHAN_ANY)

/* чтение 16 байт (2*64U) без заведения в L1 и L2 кэше */
#define __builtin_e2k_ld_128_t2(pval) __builtin_loadmas_128v (pval, 0x40, __LCC_CHAN_ANY)

/* чтение 16 байт (2*64U) без заведения во всех кэшах */
#define __builtin_e2k_ld_128_nta(pval) __builtin_loadmas_128v (pval, 0x60, __LCC_CHAN_ANY)

/* чтение 16 байт (2*64U) в спекулятивном режиме */
#define __builtin_e2k_ld_128_sm(pval, ind) __builtin_loadmas_128v ((char *) (pval) + (ind), __LCC_MAS_SPEC, __LCC_CHAN_ANY)

/* чтение 16 байт (2*64U) в спекулятивном режиме c обнулением тега */
#define __builtin_e2k_ld_128_cleartag(pval, ind) __builtin_loadmas_128v ((char *) (pval) + (ind), __LCC_MAS_SPEC | __LCC_MAS_CLEARTAG, __LCC_CHAN_ANY)

/* запись 16 байт (2*64U) в big endian режиме (с разворотом байт) */
#define __builtin_e2k_st_128_be(val, pval) __builtin_storemas_128v (val, pval, 8, __LCC_CHAN_ANY)

/* запись 16 байт (2*64U) мимо кэшей, напрямую в память (non temporal) */
#define __builtin_e2k_st_128_nt(val, pval) __builtin_storemas_128v (val, pval, 0x60, __LCC_CHAN_ANY)

/* запись 16 байт (2*64U) по маске */
#define __builtin_e2k_pst_128(val, pval, mask) \
    __builtin_pstoremas_128v (val, pval, mask, __LCC_MAS_NO, __LCC_CHAN_ANY)

/* запись 16 байт (2*64U) по маске в спекулятивном режиме */
#define __builtin_e2k_pst_128_sm(val, pval, mask) \
    __builtin_pstoremas_128v (val, pval, mask, __LCC_MAS_SPEC, __LCC_CHAN_ANY)

/* запись 16 байт (2*64U) по маске в big endian режиме (с разворотом байт) */
#define __builtin_e2k_pst_128_be(val, pval, mask) \
    __builtin_pstoremas_128v (val, pval, mask, 8, __LCC_CHAN_ANY)

#endif /* __iset__ >= 5 */

#if __iset__ >= 6 /* __elbrus_2c3__, __elbrus_12c__ и __elbrus_16c__ и старше */

/* *****************************************************************************
 *                    Новые операции в Эльбрус-2C3/12C/16C
 * *****************************************************************************
 *   Умножение вещественных чисел со сложением без промежуточного округления
 * *****************************************************************************/

/* умножение cо сложением флотов (32F) : src1 * src2 + src3
 * unsigned __builtin_e2k_fmas (unsigned, unsigned, unsigned) */

/* умножение c вычитанием флотов (32F) : src1 * src2 - src3
 * unsigned __builtin_e2k_fmss (unsigned, unsigned, unsigned) */

/* обратное умножение cо сложением флотов (32F) : -(src1 * src2) + src3
 * unsigned __builtin_e2k_fnmas (unsigned, unsigned, unsigned) */

/* обратное умножение c вычитанием флотов (32F) : -(src1 * src2) - src3
 * unsigned __builtin_e2k_fnmss (unsigned, unsigned, unsigned) */

/* умножение cо сложением даблов (64F) : src1 * src2 + src3
 * __di __builtin_e2k_fmad (__di, __di, __di) */

/* умножение c вычитанием даблов (64F) : src1 * src2 - src3
 * __di __builtin_e2k_fmsd (__di, __di, __di) */

/* обратное умножение cо сложением даблов (64F) : -(src1 * src2) + src3
 * __di __builtin_e2k_fnmad (__di, __di, __di) */

/* обратное умножение c вычитанием даблов (64F) : -(src1 * src2) - src3
 * __di __builtin_e2k_fnmsd (__di, __di, __di) */

/* *****************************************************************************
 * Упакованное умножение вещественных чисел со сложением без промежуточного округления
 * *****************************************************************************/

/* умножение 4-х флотов (32F) cо сложением : src1 * src2 + src3
 * __v2di __builtin_e2k_qpfmas (__v2di, __v2di, __v2di) */

/* умножение 4-х флотов (32F) c вычитанием : src1 * src2 - src3
 * __v2di __builtin_e2k_qpfmss (__v2di, __v2di, __v2di) */

/* обратное умножение 4-х флотов (32F) cо сложением : -(src1 * src2) + src3
 * __v2di __builtin_e2k_qpfnmas (__v2di, __v2di, __v2di) */

/* обратное умножение 4-х флотов (32F) c вычитанием : -(src1 * src2) - src3
 * __v2di __builtin_e2k_qpfnmss (__v2di, __v2di, __v2di) */

/* умножение 4-х флотов (32F) cо сложением четных элементов : src1 * src2 + src3
 * и c вычитанием нечетных элементов : src1 * src2 - src3
 * __v2di __builtin_e2k_qpfmass (__v2di, __v2di, __v2di) */

/* умножение 4-х флотов (32F) c вычитанием четных элементов : src1 * src2 - src3
 * и cо сложением нечетных элементов : src1 * src2 + src3
 * __v2di __builtin_e2k_qpfmsas (__v2di, __v2di, __v2di) */

/* умножение 2-х даблов (64F) cо сложением : src1 * src2 + src3
 * __v2di __builtin_e2k_qpfmad (__v2di, __v2di, __v2di) */

/* умножение 2-х даблов (64F) c вычитанием : src1 * src2 - src3
 * __v2di __builtin_e2k_qpfmsd (__v2di, __v2di, __v2di) */

/* обратное умножение 2-х даблов (64F) cо сложением : -(src1 * src2) + src3
 * __v2di __builtin_e2k_qpfnmad (__v2di, __v2di, __v2di) */

/* обратное умножение 2-х даблов (64F) c вычитанием : -(src1 * src2) - src3
 * __v2di __builtin_e2k_qpfnmsd (__v2di, __v2di, __v2di) */

/* умножение 2-х даблов (64F) cо сложением четных элементов : src1 * src2 + src3
 * и c вычитанием нечетных элементов : src1 * src2 - src3
 * __v2di __builtin_e2k_qpfmasd (__v2di, __v2di, __v2di) */

/* умножение 2-х даблов (64F) c вычитанием четных элементов : src1 * src2 - src3
 * и cо сложением нечетных элементов : src1 * src2 + src3
 * __v2di __builtin_e2k_qpfmsad (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                Cравнение c выработкой предиката для 64-х бит
 * *****************************************************************************/

/* сравнение на "равно" 8-ми байт (8S/8U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpeqbop (__di, __di) */

/* сравнение на "равно" 4-х шортов (16S/16U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpeqhop (__di, __di) */

/* сравнение на "равно" 2-х интов (32S/32U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpeqwop (__di, __di) */

/* сравнение на "равно" лонгов (64S/64U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpeqdop (__di, __di) */

/* сравнение на "больше" 8-ми знаковых байт (8S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpgtbop (__di, __di) */

/* сравнение на "больше" 4-х знаковых шортов (16S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpgthop (__di, __di) */

/* сравнение на "больше" 2-х знаковых интов (32S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpgtwop (__di, __di) */

/* сравнение на "больше" знаковых лонгов (64S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_pcmpgtdop (__di, __di) */

/* сравнение на "равно" 8-ми байт (8S/8U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpeqbap (__di, __di) */

/* сравнение на "равно" 4-х шортов (16S/16U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpeqhap (__di, __di) */

/* сравнение на "равно" 2-х интов (32S/32U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpeqwap (__di, __di) */

/* сравнение на "равно" лонгов (64S/64U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpeqdap (__di, __di) */

/* сравнение на "больше" 8-ми знаковых байт (8S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpgtbap (__di, __di) */

/* сравнение на "больше" 4-х знаковых шортов (16S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpgthap (__di, __di) */

/* сравнение на "больше" 2-х знаковых интов (32S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpgtwap (__di, __di) */

/* сравнение на "больше" знаковых лонгов (64S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_pcmpgtdap (__di, __di) */

/* *****************************************************************************
 *                Cравнение c выработкой предиката для 128-и бит
 * *****************************************************************************/

/* сравнение на "равно" 16-ти байт (8S/8U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqbop (__v2di, __v2di) */

/* сравнение на "равно" 8-ми шортов (16S/16U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqhop (__v2di, __v2di) */

/* сравнение на "равно" 4-х интов (32S/32U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqwop (__v2di, __v2di) */

/* сравнение на "равно" 2-х лонгов (64S/64U) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqdop (__v2di, __v2di) */

/* сравнение на "больше" 16-ти знаковых байт (8S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtbop (__v2di, __v2di) */

/* сравнение на "больше" 8-ми знаковых шортов (16S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpgthop (__v2di, __v2di) */

/* сравнение на "больше" 4-х знаковых интов (32S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtwop (__v2di, __v2di) */

/* сравнение на "больше" 2-х знаковых лонгов (64S) c последующим OR результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtdop (__v2di, __v2di) */

/* сравнение на "равно" 16-ти байт (8S/8U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqbap (__v2di, __v2di) */

/* сравнение на "равно" 8-ми шортов (16S/16U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqhap (__v2di, __v2di) */

/* сравнение на "равно" 4-х интов (32S/32U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqwap (__v2di, __v2di) */

/* сравнение на "равно" 2-х лонгов (64S/64U) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpeqdap (__v2di, __v2di) */

/* сравнение на "больше" 16-ти знаковых байт (8S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtbap (__v2di, __v2di) */

/* сравнение на "больше" 8-ми знаковых шортов (16S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpgthap (__v2di, __v2di) */

/* сравнение на "больше" 4-х знаковых интов (32S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtwap (__v2di, __v2di) */

/* сравнение на "больше" 2-х знаковых лонгов (64S) c последующим AND результатов сравнений
 * unsigned __builtin_e2k_qpcmpgtdap (__v2di, __v2di) */

/* *****************************************************************************
 *                      Беспереносное умножение целых
 * *****************************************************************************/

/* беспереносное умножение беззнаковых лонгов (64U) с выдачей младших 64-х бит результата
 * __di __builtin_e2k_clmull (__di, __di) */

/* беспереносное умножение беззнаковых лонгов (64U) с выдачей старших 64-х бит результата
 * __di __builtin_e2k_clmulh (__di, __di) */

/* *****************************************************************************
 *                               Битовый сдвиг
 * *****************************************************************************/

/* арифметический сдвиг вправо 2-х знаковых лонгов (64S)
 * __v2di __builtin_e2k_qpsrad (__v2di, __di) */

#endif /* __iset__ >= 6 */

#if __iset__ >= 7 /* __elbrus_32c__ и старше */

/* *****************************************************************************
 *                    Новые операции в Эльбрус-32C
 * *****************************************************************************/

/* Все короткие литералы в позициях src1, src2 или src3 расширяются нулем до 128 бит.
 * По крайней мере, 0 можно задавать литералом на месте любого аргумента */

/* Операция qplog добавлена в 2/5 каналы */

/* *****************************************************************************
 *          Упаковка (старые операции с новыми именами в 4-х каналах)
 * *****************************************************************************/

/* упаковка 16-ти знаковых шортов (16S) в знаковые байты (8S) с насыщением
 * __v2di __builtin_e2k_qpackhbss (__v2di, __v2di) */

/* упаковка 16-ти знаковых шортов (16S) в беззнаковые байты (8U) с насыщением
 * __v2di __builtin_e2k_qpackhbus (__v2di, __v2di) */

/* упаковка 8-ми знаковых интов (32S) в знаковые шорты (16S) с насыщением
 * __v2di __builtin_e2k_qpackwhss (__v2di, __v2di) */

/* упаковка 8-ми знаковых интов (32S) в беззнаковые шорты (16U) с насыщением
 * __v2di __builtin_e2k_qpackwhus (__v2di, __v2di) */

/* *****************************************************************************
 *                   Формирование 32 разрядной битовой маски
 * *****************************************************************************/

/* формирование маски из знака байтов
 * unsigned __builtin_e2k_qps2mb (__v2di, __v2di) */

/* формирование маски из знака шортов, каждый бит раздваивается
 * unsigned __builtin_e2k_qps2mh (__v2di, __v2di) */

/* формирование маски из знака интов, каждый бит расчетверяется
 * unsigned __builtin_e2k_qps2mw (__v2di, __v2di) */

/* формирование маски из знака лонгов, каждый бит размножается в 8
 * unsigned __builtin_e2k_qps2md (__v2di, __v2di) */

/* *****************************************************************************
 *                      Формирование сжатой битовой маски
 * *****************************************************************************/

/* формирование 16 разрядной маски из знака шортов
 * unsigned __builtin_e2k_qps2cmh (__v2di, __v2di) */

/* формирование 8 разрядной маски из знака интов
 * unsigned __builtin_e2k_qps2cmw (__v2di, __v2di) */

/* формирование 4 разрядной маски из знака лонгов
 * unsigned __builtin_e2k_qps2cmd (__v2di, __v2di) */

/* *****************************************************************************
 *             Преобразование сжатой битовой маски в полноразрядную
 * *****************************************************************************/

/* заполнение байтов битами из 32 разрядной маски
 * __v2di __builtin_e2k_qpcm2b (unsigned) */

/* заполнение шортов битами из 16 разрядной маски
 * __v2di __builtin_e2k_qpcm2h (unsigned) */

/* заполнение интов битами из 8 разрядной маски
 * __v2di __builtin_e2k_qpcm2w (unsigned) */

/* заполнение лонгов битами из 4 разрядной маски
 * __v2di __builtin_e2k_qpcm2d (unsigned) */

/* *****************************************************************************
 *                Вставление значения из операнда 2 в операнд 1
 * *****************************************************************************/

/* вставление инта
 * __v2di __builtin_e2k_qpinss (__v2di, unsigned, __v2di) */

/* вставление инта (с управляющей 32 битной константой в src3)
 * __v2di __builtin_e2k_qpinss_lt3 (__v2di, unsigned, unsigned) */

/* вставление лонга
 * __v2di __builtin_e2k_qpinsd (__v2di, __di, __v2di) */

/* вставление лонга (с управляющей 32 битной константой в src3)
 * __v2di __builtin_e2k_qpinsd_lt3 (__v2di, __di, unsigned) */

/* *****************************************************************************
 *                             Перестановка байт
 * *****************************************************************************/

/* выделение из 32-х входных байт произвольных 16-ти, положение которых и операции
 * над ними определяются управляющей 32 битной константой в src3
 * __v2di __builtin_e2k_qppermb_lt3 (__v2di, __v2di, unsigned) */

/* *****************************************************************************
 *                       Произвольные логическая операции
 * *****************************************************************************/

/* произвольная поразрядная логическая операция над 128 разрядами (128U)
 * c управляющей 32 битной константой в src3
 * __v2di __builtin_e2k_qplog_lt3 (N, __v2di, __v2di, unsigned) */

/* *****************************************************************************
 *                     Целочисленное скалярное произведение
 * *****************************************************************************/

/* 4 операции попарного умножения беззнаковых байт (8U) из src1 и src2.
 * Каждые 4-е соседних результата суммируются между собой и со знаковым интом (32S)
 * из src3. Результатом являются 4 суммы с насыщением к диапазону 32S
 * __v2di __builtin_e2k_qpidotsbwuu (__v2di, __v2di, __v2di) */

/* 4 операции попарного умножения беззнаковых байт (8U) из src1 на знаковые
 * байты (8S) из src2.
 * Каждые 4-е соседних результата суммируются между собой и со знаковым интом (32S)
 * из src3. Результатом являются 4 суммы с насыщением к диапазону 32S
 * __v2di __builtin_e2k_qpidotsbwus (__v2di, __v2di, __v2di) */

/* 4 операции попарного умножения знаковых байт (8S) из src1 и src2.
 * Каждые 4-е соседних результата суммируются между собой и со знаковым интом (32S)
 * из src3. Результатом являются 4 суммы с насыщением к диапазону 32S
 * __v2di __builtin_e2k_qpidotsbwss (__v2di, __v2di, __v2di) */

/* 4 операции попарного умножения знаковых шортов (16S) из src1 и src2.
 * Каждые два соседних результата суммируются между собой и со знаковым интом (32S)
 * из src3. Результатом являются 4 суммы с насыщением к диапазону 32S
 * __v2di __builtin_e2k_qpidotshwss (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *     Скалярное произведение над числами c плавающей точкой в формате BF16
 * *****************************************************************************/

/* 4 операции попарного умножения элементов в формате BF16 из src1 и src2
 * с результатом в формате флоат (32F).
 * Каждые два соседних результата суммируются между собой и с флотом (32F)
 * из src3. Результатом являются 4 суммы в формате флоат (32F)
 * __v2di __builtin_e2k_qpbfdots (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                          Преобразование 32F в BF16
 * *****************************************************************************/

/* преобразование 8-ми флотов (32F) в значения BF16
 * __v2di __builtin_e2k_qpfstobf (__v2di, __v2di) */

/* *****************************************************************************
 *                          Табличное преобразование байтов
 * *****************************************************************************/

/* группа (0,1,...,7) операций табличного преобразования байтов
 * __v2di __builtin_e2k_qptbl (N, __v2di, __v2di, __v2di) */

/* *****************************************************************************
 *                 Операции поддержки алгоритма шифрования AES
 * *****************************************************************************/

/* начало раунда AES шифрования
 * __v2di __builtin_e2k_qpaesebgn (__v2di, __v2di, __v2di) */

/* линейное преобразование AES шифрования
 * __v2di __builtin_e2k_qpaesltr (__v2di, __v2di, __v2di) */

/* начало раунда AES дешифрования
 * __v2di __builtin_e2k_qpaesdbgn (__v2di, __v2di, __v2di) */

/* линейное преобразование AES дешифрования
 * __v2di __builtin_e2k_qpaesiltr (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *       Операции поддержки алгоритма шифрования Кузнечик (ГОСТ 34.12-2018)
 * *****************************************************************************/

/* начало раунда шифрования алгоритма Кузнечик
 * __v2di __builtin_e2k_qpkuzebgn (__v2di, __v2di, __v2di) */

/* линейное преобразование алгоритма шифрования Кузнечик
 * __v2di __builtin_e2k_qpkuzltr (__v2di) */

/* *****************************************************************************
 *       Операция поддержки алгоритма шифрования Магма (ГОСТ 34.12-2018)
 * *****************************************************************************/

/* раунд шифрования алгоритма Магма для 4-х блоков
 * __v2di __builtin_e2k_qpmgme (__v2di, __v2di, __v2di) */

/* *****************************************************************************
 *       Операция поддержки алгоритма хэширования Стрибог (ГОСТ 34.11-2018)
 * *****************************************************************************/

/* L-преобразование алгоритма хэширования Стрибог для части хэшируемого блока
 * __v2di __builtin_e2k_qpsbgltr (__v2di) */

#endif /* __iset__ >= 7 */

#undef TWO_OPERANDS_INSTRUCTIONS
#undef TWO_OPERANDS_INSTRUCTIONS_QSQ

/* *****************************************************************************
 *                                   Макросы
 * *****************************************************************************/

/* *****************************************************************************
 *                       Выравнивание данных в Эльбрус
 *
 *    Используется, например, для выравнивания данных, считанных операциями
 *    mova(d,w), работающими только с выровненными адресами. Или, например, для
 *    сдвига на заранее неизвестное число байт
 *
 *    Примеры использования можно найти в <e2kbuiltin.h>
 * *****************************************************************************/

#define E2K_BYTES_TO_ALIGN(ptr, align) ((-(long)(ptr))&((align)-1))
#define E2K_BYTES_FROM_ALIGN(ptr, align) (((long)(ptr))&((align)-1))

#ifdef __ptr128__
#define E2K_ALIGN_PTR(ptr, align) \
    __builtin_assume_aligned ((char*)(ptr) + (E2K_BYTES_TO_ALIGN(ptr, align)), align)
#define E2K_ALIGN_PTR_BACK(ptr, align) \
    __builtin_assume_aligned ((char*)(ptr) - (E2K_BYTES_FROM_ALIGN(ptr, align)), align)
#else /* __ptr128__ */
#define E2K_ALIGN_PTR(ptr, align) \
    __builtin_assume_aligned ((void*) (((long)(ptr) + ((align)-1)) & ~((align)-1)), align)
#define E2K_ALIGN_PTR_BACK(ptr, align) \
    __builtin_assume_aligned ((void*) ((long)(ptr) & ~((align)-1)), align)
#endif /* __ptr128__ */

/* вариант на арифметических операциях для данных формата 64 */
#define E2K_PREPARE_ALIGN_ARITH(pointer, align)           \
    align <<= 3
#define E2K_ALIGN_DATA_ARITH(src1, src2, dst, align)      \
    dst = (((__di)(src1)) >> align) | (((src2) << (64 - align)) & ((- (long long) align) >> 63))

/* вариант на арифметических операциях для данных формата 64, считанных в big-endian режиме */
#define E2K_PREPARE_ALIGN_ARITH_BE(pointer, align)        \
    E2K_PREPARE_ALIGN_ARITH(pointer, align)
#define E2K_ALIGN_DATA_ARITH_BE(src1, src2, dst, align)   \
    dst = ((src1) << align) | ((((__di)(src2)) >> (64 - align)) & ((- (long long) align) >> 63))

/* вариант на арифметических операциях для данных формата 32 */
#define E2K_PREPARE_ALIGN_ARITH32(pointer, align)         \
    align <<= 3
#define E2K_ALIGN_DATA_ARITH32(src1, src2, dst, align)    \
    dst = ((__di) (src1) >> align) | (((__di) (src2) << (32 - align)) & ((- (long long) align) >> 31))

/* вариант на операциях insfs/scrs для данных формата 32 */
#define E2K_PREPARE_ALIGN_INSF32(pointer, align)          \
    align <<= 9
#define E2K_ALIGN_DATA_INSF32(src1, src2, dst, align)     \
    dst = __builtin_e2k_scrs (__builtin_e2k_insfs (src1, align, src2), align >> 6)

/* вариант на операции pshufb для данных формата 64 */
#define E2K_PREPARE_ALIGN(align, spec)                    \
    spec = __builtin_e2k_paddd (__builtin_e2k_pshufb (0, align, 0), 0x0706050403020100LL)
#define E2K_ALIGN_DATA(src1, src2, dst, spec)             \
    dst = __builtin_e2k_pshufb (src2, src1, spec)

/* вариант на операции pshufb для данных формата 64, считанных в big-endian режиме */
#define E2K_PREPARE_ALIGN_BE(align, spec)                 \
    spec = __builtin_e2k_psubd (0x0f0e0d0c0b0a0908LL, __builtin_e2k_pshufb (0, align, 0))
#define E2K_ALIGN_DATA_BE(src1, src2, dst, spec)          \
    dst = __builtin_e2k_pshufb (src1, src2, spec)

#if __iset__ >= 5 /* __elbrus_8c2__ и старше */

/* вариант на операции qppermb для данных формата 128 */
#define E2K_PREPARE_ALIGN128(align, spec) {                                               \
    __di __xx = __builtin_e2k_pshufb (0, align, 0);                                       \
    spec = __builtin_e2k_qppackdl (__builtin_e2k_paddd (__xx, 0x0f0e0d0c0b0a0908LL),      \
                                   __builtin_e2k_paddd (__xx, 0x0706050403020100LL));     \
}
#define E2K_ALIGN_DATA128(src1, src2, dst, spec)                                          \
    dst = __builtin_e2k_qppermb (src2, src1, spec)

/* вариант на операции qppermb для данных формата 128, считанных в big-endian режиме */
#define E2K_PREPARE_ALIGN128_BE(align, spec) {                                            \
    __di __xx = __builtin_e2k_pshufb (0, align, 0);                                       \
    spec = __builtin_e2k_qppackdl (__builtin_e2k_psubd (0x1f1e1d1c1b1a1918LL, __xx),      \
                                   __builtin_e2k_psubd (0x1716151413121110LL, __xx));     \
}
#define E2K_ALIGN_DATA128_BE(src1, src2, dst, spec)                                       \
    dst = __builtin_e2k_qppermb (src1, src2, spec)

#endif /* __iset__ >= 5 */


/* *****************************************************************************
 *       Вычисление номера первого ненулевого байта в 64S/64U
 * *****************************************************************************/

#define __CTZB(count, x) /* count trailng zero bytes */   \
    (count) = __builtin_e2k_lzcntd (__builtin_e2k_bitrevd (x)) >> 3

#define __CLZB(count, x) /* count leading zero bytes */   \
    (count) = __builtin_e2k_lzcntd (x) >> 3

/* *****************************************************************************
 *                       Размеры L1/L2/L3 кэшей в Эльбрус
 * *****************************************************************************/

#define SIZE_L1 65536            /* размер L1 кэша в байтах */

#if __iset__ <= 2 /* __elbrus_2cplus__ */

#define SIZE_L2 1048576          /* размер L2 кэша в байтах (e3s) */

#elif __iset__ <= 3 || (__iset__ == 4 && defined (__elbrus_1cplus__) && !defined (__elbrus_8c__)) || \
     (__iset__ == 6 && defined (__elbrus_2c3__)) /* __elbrus_4c__, __elbrus_1cplus__, __elbrus_2c3__ */

#define SIZE_L2 2097152          /* размер L2 кэша в байтах */

#elif __iset__ <= 5 /* __elbrus_8c__ и __elbrus_8c2__ */

#define SIZE_L2 524288           /* размер L2 кэша в байтах */
#define SIZE_L3 16777216         /* размер L3 кэша в байтах */

#elif __iset__ <= 6 /* __elbrus_12c__ и __elbrus_16c__ */

#define SIZE_L2 1048576          /* размер L2 кэша в байтах */
#define SIZE_L3 33554432         /* размер L3 кэша в байтах */

#elif __iset__ <= 7 /* __elbrus_32c__ */

#define SIZE_L2 1048576          /* размер L2 кэша в байтах */
#define SIZE_L3 50331648         /* размер L3 кэша в байтах */

#endif /* __iset__ <= 7 */

#ifdef __cplusplus
}
#endif

#endif /* _E2KINTRIN_H */
